# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests of bound branches (binding, unbinding, commit, etc) command."""

from breezy import branch, controldir, errors, tests
from breezy.tests import script


class TestBoundBranches(tests.TestCaseWithTransport):
    def create_branches(self):
        base_tree = self.make_branch_and_tree("base")
        base_tree.lock_write()
        self.build_tree(["base/a", "base/b"])
        base_tree.add(["a", "b"])
        base_tree.commit("init")
        base_tree.unlock()

        child_tree = base_tree.branch.create_checkout("child")

        self.check_revno(1, "child")
        d = controldir.ControlDir.open("child")
        self.assertNotEqual(None, d.open_branch().get_master_branch())

        return base_tree, child_tree

    def check_revno(self, val, loc="."):
        self.assertEqual(
            val, controldir.ControlDir.open(loc).open_branch().last_revision_info()[0]
        )

    def test_simple_binding(self):
        tree = self.make_branch_and_tree("base")
        self.build_tree(["base/a", "base/b"])
        tree.add("a", ids=b"b")
        tree.commit(message="init")

        tree.controldir.sprout("child")

        self.run_bzr("bind ../base", working_dir="child")

        d = controldir.ControlDir.open("child")
        self.assertNotEqual(None, d.open_branch().get_master_branch())

        self.run_bzr("unbind", working_dir="child")
        self.assertEqual(None, d.open_branch().get_master_branch())

        self.run_bzr("unbind", retcode=3, working_dir="child")

    def test_bind_branch6(self):
        self.make_branch("branch1", format="dirstate-tags")
        error = self.run_bzr("bind", retcode=3, working_dir="branch1")[1]
        self.assertEndsWith(
            error, "No location supplied and no previous location known\n"
        )

    def setup_rebind(self, format):
        branch1 = self.make_branch("branch1")
        branch2 = self.make_branch("branch2", format=format)
        branch2.bind(branch1)
        branch2.unbind()

    def test_rebind_branch6(self):
        self.setup_rebind("dirstate-tags")
        self.run_bzr("bind", working_dir="branch2")
        b = branch.Branch.open("branch2")
        self.assertEndsWith(b.get_bound_location(), "/branch1/")

    def test_rebind_branch5(self):
        self.setup_rebind("knit")
        error = self.run_bzr("bind", retcode=3, working_dir="branch2")[1]
        self.assertEndsWith(
            error,
            "No location supplied.  This format does not remember old locations.\n",
        )

    def test_bound_commit(self):
        child_tree = self.create_branches()[1]

        self.build_tree_contents([("child/a", b"new contents")])
        child_tree.commit(message="child")

        self.check_revno(2, "child")

        # Make sure it committed on the parent
        self.check_revno(2, "base")

    def test_bound_fail(self):
        # Make sure commit fails if out of date.
        base_tree, child_tree = self.create_branches()

        self.build_tree_contents(
            [("base/a", b"new base contents\n"), ("child/b", b"new b child contents\n")]
        )
        base_tree.commit(message="base")
        self.check_revno(2, "base")

        self.check_revno(1, "child")
        self.assertRaises(
            errors.BoundBranchOutOfDate, child_tree.commit, message="child"
        )
        self.check_revno(1, "child")

        child_tree.update()
        self.check_revno(2, "child")

        child_tree.commit(message="child")
        self.check_revno(3, "child")
        self.check_revno(3, "base")

    def test_double_binding(self):
        child_tree = self.create_branches()[1]
        child_tree.controldir.sprout("child2")

        # Double binding succeeds, but committing to child2 should fail
        self.run_bzr("bind ../child", working_dir="child2")

        # Refresh the child tree object as 'unbind' modified it
        child2_tree = controldir.ControlDir.open("child2").open_workingtree()
        self.assertRaises(
            errors.CommitToDoubleBoundBranch,
            child2_tree.commit,
            message="child2",
            allow_pointless=True,
        )

    def test_unbinding(self):
        base_tree, child_tree = self.create_branches()

        self.build_tree_contents(
            [("base/a", b"new base contents\n"), ("child/b", b"new b child contents\n")]
        )

        base_tree.commit(message="base")
        self.check_revno(2, "base")

        self.check_revno(1, "child")
        self.run_bzr("commit -m child", retcode=3, working_dir="child")
        self.check_revno(1, "child")
        self.run_bzr("unbind", working_dir="child")
        # Refresh the child tree/branch objects as 'unbind' modified them
        child_tree = child_tree.controldir.open_workingtree()
        child_tree.commit(message="child")
        self.check_revno(2, "child")

    def test_commit_remote_bound(self):
        # It is not possible to commit to a branch
        # which is bound to a branch which is bound
        base_tree, child_tree = self.create_branches()
        base_tree.controldir.sprout("newbase")

        # There is no way to know that B has already
        # been bound by someone else, otherwise it
        # might be nice if this would fail
        self.run_bzr("bind ../newbase", working_dir="base")

        self.run_bzr("commit -m failure --unchanged", retcode=3, working_dir="child")

    def test_pull_updates_both(self):
        base_tree = self.create_branches()[0]
        newchild_tree = base_tree.controldir.sprout("newchild").open_workingtree()
        self.build_tree_contents([("newchild/b", b"newchild b contents\n")])
        newchild_tree.commit(message="newchild")
        self.check_revno(2, "newchild")

        # The pull should succeed, and update
        # the bound parent branch
        self.run_bzr("pull ../newchild", working_dir="child")
        self.check_revno(2, "child")
        self.check_revno(2, "base")

    def test_pull_local_updates_local(self):
        base_tree = self.create_branches()[0]
        newchild_tree = base_tree.controldir.sprout("newchild").open_workingtree()
        self.build_tree_contents([("newchild/b", b"newchild b contents\n")])
        newchild_tree.commit(message="newchild")
        self.check_revno(2, "newchild")

        # The pull should succeed, and update
        # the bound parent branch
        self.run_bzr("pull ../newchild --local", working_dir="child")
        self.check_revno(2, "child")
        self.check_revno(1, "base")

    def test_bind_diverged(self):
        base_tree, child_tree = self.create_branches()

        self.run_bzr("unbind", working_dir="child")

        # Refresh the child tree/branch objects as 'unbind' modified them
        child_tree = child_tree.controldir.open_workingtree()
        child_tree.commit(message="child", allow_pointless=True)
        self.check_revno(2, "child")

        self.check_revno(1, "base")
        base_tree.commit(message="base", allow_pointless=True)
        self.check_revno(2, "base")

        # These branches have diverged, but bind should succeed anyway
        self.run_bzr("bind ../base", working_dir="child")

        # Refresh the child tree/branch objects as 'bind' modified them
        child_tree = child_tree.controldir.open_workingtree()
        # This should turn the local commit into a merge
        child_tree.update()
        child_tree.commit(message="merged")
        self.check_revno(3, "child")
        self.assertEqual(
            child_tree.branch.last_revision(), base_tree.branch.last_revision()
        )

    def test_bind_parent_ahead(self):
        base_tree = self.create_branches()[0]

        self.run_bzr("unbind", working_dir="child")

        base_tree.commit(message="base", allow_pointless=True)

        self.check_revno(1, "child")
        self.run_bzr("bind ../base", working_dir="child")

        # binding does not pull data:
        self.check_revno(1, "child")
        self.run_bzr("unbind", working_dir="child")

        # Check and make sure it also works if parent is ahead multiple
        base_tree.commit(message="base 3", allow_pointless=True)
        base_tree.commit(message="base 4", allow_pointless=True)
        base_tree.commit(message="base 5", allow_pointless=True)
        self.check_revno(5, "base")

        self.check_revno(1, "child")
        self.run_bzr("bind ../base", working_dir="child")
        self.check_revno(1, "child")

    def test_bind_child_ahead(self):
        # test binding when the master branches history is a prefix of the
        # childs - it should bind ok but the revision histories should not
        # be altered
        child_tree = self.create_branches()[1]

        self.run_bzr("unbind", working_dir="child")
        # Refresh the child tree/branch objects as 'bind' modified them
        child_tree = child_tree.controldir.open_workingtree()
        child_tree.commit(message="child", allow_pointless=True)
        self.check_revno(2, "child")
        self.check_revno(1, "base")

        self.run_bzr("bind ../base", working_dir="child")
        self.check_revno(1, "base")

        # Check and make sure it also works if child is ahead multiple
        self.run_bzr("unbind", working_dir="child")
        child_tree.commit(message="child 3", allow_pointless=True)
        child_tree.commit(message="child 4", allow_pointless=True)
        child_tree.commit(message="child 5", allow_pointless=True)
        self.check_revno(5, "child")

        self.check_revno(1, "base")
        self.run_bzr("bind ../base", working_dir="child")
        self.check_revno(1, "base")

    def test_bind_fail_if_missing(self):
        """We should not be able to bind to a missing branch."""
        tree = self.make_branch_and_tree("tree_1")
        tree.commit("dummy commit")
        self.run_bzr_error(
            ["Not a branch.*no-such-branch/"],
            ["bind", "../no-such-branch"],
            working_dir="tree_1",
        )
        self.assertIs(None, tree.branch.get_bound_location())

    def test_commit_after_merge(self):
        base_tree, child_tree = self.create_branches()

        # We want merge to be able to be a local only
        # operation, because it can be without violating
        # the binding invariants.
        # But we can't fail afterwards
        other_tree = child_tree.controldir.sprout("other").open_workingtree()
        other_branch = other_tree.branch

        self.build_tree_contents([("other/c", b"file c\n")])
        other_tree.add("c")
        other_tree.commit(message="adding c")
        new_rev_id = other_branch.last_revision()

        child_tree.merge_from_branch(other_branch)

        self.assertPathExists("child/c")
        self.assertEqual([new_rev_id], child_tree.get_parent_ids()[1:])

        # Make sure the local branch has the installed revision
        self.assertTrue(child_tree.branch.repository.has_revision(new_rev_id))

        # And make sure that the base tree does not
        self.assertFalse(base_tree.branch.repository.has_revision(new_rev_id))

        # Commit should succeed, and cause merged revisions to
        # be pulled into base
        self.run_bzr(["commit", "-m", "merge other"], working_dir="child")
        self.check_revno(2, "child")
        self.check_revno(2, "base")
        self.assertTrue(base_tree.branch.repository.has_revision(new_rev_id))

    def test_pull_overwrite(self):
        # XXX: This test should be moved to branch-implemenations/test_pull
        child_tree = self.create_branches()[1]

        other_tree = child_tree.controldir.sprout("other").open_workingtree()

        self.build_tree_contents([("other/a", b"new contents\n")])
        other_tree.commit(message="changed a")
        self.check_revno(2, "other")
        self.build_tree_contents([("other/a", b"new contents\nand then some\n")])
        other_tree.commit(message="another a")
        self.check_revno(3, "other")
        self.build_tree_contents(
            [("other/a", b"new contents\nand then some\nand some more\n")]
        )
        other_tree.commit("yet another a")
        self.check_revno(4, "other")

        self.build_tree_contents([("child/a", b"also changed a\n")])
        child_tree.commit(message="child modified a")

        self.check_revno(2, "child")
        self.check_revno(2, "base")

        self.run_bzr("pull --overwrite ../other", working_dir="child")

        # both the local and master should have been updated.
        self.check_revno(4, "child")
        self.check_revno(4, "base")

    def test_bind_directory(self):
        """Test --directory option."""
        tree = self.make_branch_and_tree("base")
        self.build_tree(["base/a", "base/b"])
        tree.add("a", ids=b"b")
        tree.commit(message="init")
        tree.controldir.sprout("child")
        self.run_bzr("bind --directory=child base")
        d = controldir.ControlDir.open("child")
        self.assertNotEqual(None, d.open_branch().get_master_branch())
        self.run_bzr("unbind -d child")
        self.assertEqual(None, d.open_branch().get_master_branch())
        self.run_bzr("unbind --directory child", retcode=3)


class TestBind(script.TestCaseWithTransportAndScript):
    def test_bind_when_bound(self):
        self.run_script("""
$ brz init trunk
...
$ brz init copy
...
$ cd copy
$ brz bind ../trunk
$ brz bind
2>brz: ERROR: Branch is already bound
""")

    def test_bind_before_bound(self):
        self.run_script("""
$ brz init trunk
...
$ cd trunk
$ brz bind
2>brz: ERROR: No location supplied and no previous location known
""")



# Copyright (C) 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that we can use smart_add on all Tree implementations."""

import os
import sys
from io import StringIO

from ... import add as _mod_add
from ... import errors, ignores, osutils, tests, trace, transport, workingtree
from .. import features, per_workingtree, test_smart_add


class RecordingAddAction(_mod_add.AddAction):
    def __init__(self):
        self.adds = []

    def __call__(self, wt, parent_ie, path, kind):
        self.adds.append((wt, path, kind))


class TestSmartAddTree(per_workingtree.TestCaseWithWorkingTree):
    def test_single_file(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a"])
        action = RecordingAddAction()
        tree.smart_add(["tree"], action=action)

        with tree.lock_read():
            files = [
                (path, status, kind)
                for path, status, kind, parent_id in tree.list_files(include_root=True)
            ]
        self.assertEqual([("", "V", "directory"), ("a", "V", "file")], files)
        self.assertEqual([(tree, "a", "file")], action.adds)

    def assertFilenameSkipped(self, filename):
        tree = self.make_branch_and_tree("tree")
        try:
            self.build_tree(["tree/" + filename])
        except transport.NoSuchFile:
            if sys.platform == "win32":
                raise tests.TestNotApplicable(
                    "Cannot create files named {!r} on win32".format(filename)
                )
        tree.smart_add(["tree"])
        self.assertFalse(tree.is_versioned(filename))

    def test_path_containing_newline_skips(self):
        self.assertFilenameSkipped("a\nb")

    def test_path_containing_carriagereturn_skips(self):
        self.assertFilenameSkipped("a\rb")

    def test_save_false(self):
        """Dry-run add doesn't permanently affect the tree."""
        wt = self.make_branch_and_tree(".")
        with wt.lock_write():
            self.build_tree(["file"])
            wt.smart_add(["file"], save=False)
            # the file should not be added - no id.
            self.assertFalse(wt.is_versioned("file"))
        # and the disk state should be the same - reopen to check.
        wt = wt.controldir.open_workingtree()
        self.assertFalse(wt.is_versioned("file"))

    def test_add_dot_from_root(self):
        """Test adding . from the root of the tree."""
        paths = ("original/", "original/file1", "original/file2")
        self.build_tree(paths)
        wt = self.make_branch_and_tree(".")
        action = RecordingAddAction()
        wt.smart_add((".",), action=action)
        for path in paths:
            self.assertTrue(wt.is_versioned(path))
        if wt.has_versioned_directories():
            self.assertEqual(
                {
                    (wt, "original", "directory"),
                    (wt, "original/file1", "file"),
                    (wt, "original/file2", "file"),
                },
                set(action.adds),
            )
        else:
            self.assertEqual(
                {(wt, "original/file1", "file"), (wt, "original/file2", "file")},
                set(action.adds),
            )

    def test_skip_nested_trees(self):
        """Test smart-adding a nested tree ignors it and warns."""
        wt = self.make_branch_and_tree(".")
        nested_wt = self.make_branch_and_tree("nested")
        warnings = []

        def warning(*args):
            warnings.append(args[0] % args[1:])

        self.overrideAttr(trace, "warning", warning)
        wt.smart_add((".",))
        self.assertFalse(wt.is_versioned("nested"))
        self.assertEqual(
            ["skipping nested tree {!r}".format(nested_wt.basedir)], warnings
        )

    def test_add_dot_from_subdir(self):
        """Test adding . from a subdir of the tree."""
        paths = ("original/", "original/file1", "original/file2")
        self.build_tree(paths)
        wt = self.make_branch_and_tree(".")
        wt.smart_add((".",))
        for path in paths:
            self.assertTrue(wt.is_versioned(path))

    def test_add_tree_from_above_tree(self):
        """Test adding a tree from above the tree."""
        paths = ("original/", "original/file1", "original/file2")
        branch_paths = (
            "branch/",
            "branch/original/",
            "branch/original/file1",
            "branch/original/file2",
        )
        self.build_tree(branch_paths)
        wt = self.make_branch_and_tree("branch")
        wt.smart_add(("branch",))
        for path in paths:
            self.assertTrue(wt.is_versioned(path))

    def test_add_above_tree_preserves_tree(self):
        """Test nested trees are not affect by an add above them."""
        paths = ("original/", "original/file1", "original/file2")
        child_paths = ("path",)
        full_child_paths = ("original/child", "original/child/path")
        build_paths = (
            "original/",
            "original/file1",
            "original/file2",
            "original/child/",
            "original/child/path",
        )

        self.build_tree(build_paths)
        wt = self.make_branch_and_tree(".")
        if wt.controldir.user_url != wt.branch.controldir.user_url:
            # Lightweight checkout, make sure we have a repo location.
            wt.branch.controldir.root_transport.mkdir("original")
        child_tree = self.make_branch_and_tree("original/child")
        wt.smart_add((".",))
        for path in paths:
            self.assertNotEqual((path, wt.is_versioned(path)), (path, False))
        for path in full_child_paths:
            self.assertEqual((path, wt.is_versioned(path)), (path, False))
        for path in child_paths:
            self.assertFalse(child_tree.is_versioned(path))

    def test_add_paths(self):
        """Test smart-adding a list of paths."""
        paths = ("file1", "file2")
        self.build_tree(paths)
        wt = self.make_branch_and_tree(".")
        wt.smart_add(paths)
        for path in paths:
            self.assertTrue(wt.is_versioned(path))

    def test_add_ignored_nested_paths(self):
        """Test smart-adding a list of paths which includes ignored ones."""
        wt = self.make_branch_and_tree(".")
        tree_shape = ("adir/", "adir/CVS/", "adir/CVS/afile", "adir/CVS/afile2")
        add_paths = ("adir/CVS", "adir/CVS/afile", "adir")
        expected_paths = ("adir", "adir/CVS", "adir/CVS/afile", "adir/CVS/afile2")
        self.build_tree(tree_shape)
        wt.smart_add(add_paths)
        for path in expected_paths:
            self.assertTrue(wt.is_versioned(path), "No id added for {}".format(path))

    def test_add_non_existant(self):
        """Test smart-adding a file that does not exist."""
        wt = self.make_branch_and_tree(".")
        self.assertRaises(transport.NoSuchFile, wt.smart_add, ["non-existant-file"])

    def test_returns_and_ignores(self):
        """Correctly returns added/ignored files."""
        wt = self.make_branch_and_tree(".")
        # The default ignore list includes '*.py[co]', but not CVS
        ignores._set_user_ignores(["*.py[co]"])
        self.build_tree(
            ["inertiatic/", "inertiatic/esp", "inertiatic/CVS", "inertiatic/foo.pyc"]
        )
        added, ignored = wt.smart_add(".")
        if wt.has_versioned_directories():
            self.assertSubset(("inertiatic", "inertiatic/esp", "inertiatic/CVS"), added)
        else:
            self.assertSubset(("inertiatic/esp", "inertiatic/CVS"), added)
        self.assertSubset(("*.py[co]",), ignored)
        self.assertSubset(("inertiatic/foo.pyc",), ignored["*.py[co]"])

    def test_add_multiple_dirs(self):
        """Test smart adding multiple directories at once."""
        added_paths = [
            "file1",
            "file2",
            "dir1/",
            "dir1/file3",
            "dir1/subdir2/",
            "dir1/subdir2/file4",
            "dir2/",
            "dir2/file5",
        ]
        not_added = ["file6", "dir3/", "dir3/file7", "dir3/file8"]
        self.build_tree(added_paths)
        self.build_tree(not_added)

        wt = self.make_branch_and_tree(".")
        wt.smart_add(["file1", "file2", "dir1", "dir2"])

        for path in added_paths:
            self.assertTrue(
                wt.is_versioned(path.rstrip("/")), "Failed to add path: {}".format(path)
            )
        for path in not_added:
            self.assertFalse(
                wt.is_versioned(path.rstrip("/")),
                "Accidentally added path: {}".format(path),
            )

    def test_add_file_in_unknown_dir(self):
        # Test that parent directory addition is implicit
        tree = self.make_branch_and_tree(".")
        self.build_tree(["dir/", "dir/subdir/", "dir/subdir/foo"])
        tree.smart_add(["dir/subdir/foo"])
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            ["", "dir", "dir/subdir", "dir/subdir/foo"],
            [path for path, ie in tree.iter_entries_by_dir()],
        )

    def test_add_dir_bug_251864(self):
        """Added file turning into a dir should be detected on add dir.

        Similar to bug 205636 but with automatic adding of directory contents.
        """
        tree = self.make_branch_and_tree(".")
        self.build_tree(["dir"])  # whoops, make a file called dir
        tree.smart_add(["dir"])
        os.remove("dir")
        self.build_tree(["dir/", "dir/file"])
        tree.smart_add(["dir"])
        tree.commit("Add dir contents")
        self.addCleanup(tree.lock_read().unlock)
        self.assertEqual(
            [("dir", "directory"), ("dir/file", "file")],
            [(t[0], t[2]) for t in tree.list_files()],
        )
        self.assertFalse(list(tree.iter_changes(tree.basis_tree())))

    def test_add_subdir_file_bug_205636(self):
        """Added file turning into a dir should be detected on add dir/file."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["dir"])  # whoops, make a file called dir
        tree.smart_add(["dir"])
        os.remove("dir")
        self.build_tree(["dir/", "dir/file"])
        tree.smart_add(["dir/file"])
        tree.commit("Add file in dir")
        self.addCleanup(tree.lock_read().unlock)
        self.assertEqual(
            [("dir", "directory"), ("dir/file", "file")],
            [(t[0], t[2]) for t in tree.list_files()],
        )
        self.assertFalse(list(tree.iter_changes(tree.basis_tree())))

    def test_custom_ids(self):
        sio = StringIO()
        action = test_smart_add.AddCustomIDAction(to_file=sio, should_print=True)
        self.build_tree(["file1", "dir1/", "dir1/file2"])

        wt = self.make_branch_and_tree(".")
        if not wt._format.supports_setting_file_ids:
            self.assertRaises(
                workingtree.SettingFileIdUnsupported, wt.smart_add, ["."], action=action
            )
            return

        wt.smart_add(["."], action=action)
        # The order of adds is not strictly fixed:
        sio.seek(0)
        lines = sorted(sio.readlines())
        self.assertEqual(
            [
                "added dir1 with id directory-dir1\n",
                "added dir1/file2 with id file-dir1%file2\n",
                "added file1 with id file-file1\n",
            ],
            lines,
        )
        wt.lock_read()
        self.addCleanup(wt.unlock)
        self.assertEqual(
            [
                ("", wt.path2id("")),
                ("dir1", b"directory-dir1"),
                ("file1", b"file-file1"),
                ("dir1/file2", b"file-dir1%file2"),
            ],
            [(path, ie.file_id) for path, ie in wt.iter_entries_by_dir()],
        )


class TestSmartAddConflictRelatedFiles(per_workingtree.TestCaseWithWorkingTree):
    def make_tree_with_text_conflict(self):
        tb = self.make_branch_and_tree("base")
        self.build_tree_contents([("base/file", b"content in base")])
        tb.add("file")
        tb.commit("Adding file")

        t1 = tb.controldir.sprout("t1").open_workingtree()

        self.build_tree_contents([("base/file", b"content changed in base")])
        tb.commit("Changing file in base")

        self.build_tree_contents([("t1/file", b"content in t1")])
        t1.commit("Changing file in t1")
        t1.merge_from_branch(tb.branch)
        fnames = ["file.{}".format(s) for s in ("BASE", "THIS", "OTHER")]
        for fn in fnames:
            self.assertPathExists(os.path.join(t1.basedir, fn))
        return t1

    def test_cant_add_generated_files_implicitly(self):
        t = self.make_tree_with_text_conflict()
        added, ignored = t.smart_add([t.basedir])
        self.assertEqual(([], {}), (added, ignored))

    def test_can_add_generated_files_explicitly(self):
        fnames = ["file.{}".format(s) for s in ("BASE", "THIS", "OTHER")]
        t = self.make_tree_with_text_conflict()
        added, ignored = t.smart_add([t.basedir + "/{}".format(f) for f in fnames])
        self.assertEqual((fnames, {}), (added, ignored))


class TestSmartAddTreeUnicode(per_workingtree.TestCaseWithWorkingTree):
    _test_needs_features = [features.UnicodeFilenameFeature]

    def setUp(self):
        super().setUp()
        self.build_tree(["a\u030a"])
        self.wt = self.make_branch_and_tree(".")
        self.overrideAttr(osutils, "normalized_filename")

    def test_requires_normalized_unicode_filenames_fails_on_unnormalized(self):
        """Adding unnormalized unicode filenames fail if and only if the
        workingtree format has the requires_normalized_unicode_filenames flag
        set and the underlying filesystem doesn't normalize.
        """
        osutils.normalized_filename = osutils._accessible_normalized_filename
        if (
            self.workingtree_format.requires_normalized_unicode_filenames
            and sys.platform != "darwin"
        ):
            self.assertRaises(transport.NoSuchFile, self.wt.smart_add, ["a\u030a"])
        else:
            self.wt.smart_add(["a\u030a"])

    def test_accessible_explicit(self):
        osutils.normalized_filename = osutils._accessible_normalized_filename
        if self.workingtree_format.requires_normalized_unicode_filenames:
            raise tests.TestNotApplicable(
                "Working tree format smart_add requires normalized unicode filenames"
            )
        self.wt.smart_add(["a\u030a"])
        self.wt.lock_read()
        self.addCleanup(self.wt.unlock)
        self.assertEqual(
            [("", "directory"), ("\xe5", "file")],
            [(path, ie.kind) for path, ie in self.wt.iter_entries_by_dir()],
        )

    def test_accessible_implicit(self):
        osutils.normalized_filename = osutils._accessible_normalized_filename
        if self.workingtree_format.requires_normalized_unicode_filenames:
            raise tests.TestNotApplicable(
                "Working tree format smart_add requires normalized unicode filenames"
            )
        self.wt.smart_add([])
        self.wt.lock_read()
        self.addCleanup(self.wt.unlock)
        self.assertEqual(
            [("", "directory"), ("\xe5", "file")],
            [(path, ie.kind) for path, ie in self.wt.iter_entries_by_dir()],
        )

    def test_inaccessible_explicit(self):
        osutils.normalized_filename = osutils._inaccessible_normalized_filename
        self.assertRaises(errors.InvalidNormalization, self.wt.smart_add, ["a\u030a"])

    def test_inaccessible_implicit(self):
        osutils.normalized_filename = osutils._inaccessible_normalized_filename
        # TODO: jam 20060701 In the future, this should probably
        #       just ignore files that don't fit the normalization
        #       rules, rather than exploding
        self.assertRaises(errors.InvalidNormalization, self.wt.smart_add, [])



# Copyright (C) 2006, 2007, 2009-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os

from breezy.tests import TestCaseWithTransport
from breezy.workingtree import WorkingTree


class TestRemerge(TestCaseWithTransport):
    def make_file(self, name, contents):
        with open(name, "w") as f:
            f.write(contents)

    def create_conflicts(self):
        """Create a conflicted tree."""
        os.mkdir("base")
        self.make_file("base/hello", "hi world")
        self.make_file("base/answer", "42")
        self.run_bzr("init", working_dir="base")
        self.run_bzr("add", working_dir="base")
        self.run_bzr("commit -m base", working_dir="base")
        self.run_bzr("branch base other")
        self.run_bzr("branch base this")
        self.make_file("other/hello", "Hello.")
        self.make_file("other/answer", "Is anyone there?")
        self.run_bzr("commit -m other", working_dir="other")
        self.make_file("this/hello", "Hello, world")
        self.run_bzr("mv answer question", working_dir="this")
        self.make_file(
            "this/question", "What do you get when you multiply sixtimes nine?"
        )
        self.run_bzr("commit -m this", working_dir="this")

    def test_remerge(self):
        """Remerge command works as expected."""
        self.create_conflicts()
        self.run_bzr("merge ../other --show-base", retcode=1, working_dir="this")
        with open("this/hello") as f:
            conflict_text = f.read()
        self.assertTrue("|||||||" in conflict_text)
        self.assertTrue("hi world" in conflict_text)

        self.run_bzr_error(
            ["conflicts encountered"], "remerge", retcode=1, working_dir="this"
        )
        with open("this/hello") as f:
            conflict_text = f.read()
        self.assertFalse("|||||||" in conflict_text)
        self.assertFalse("hi world" in conflict_text)

        os.unlink("this/hello.OTHER")
        os.unlink("this/question.OTHER")

        self.run_bzr_error(
            ["jello is not versioned"],
            "remerge jello --merge-type weave",
            working_dir="this",
        )
        self.run_bzr_error(
            ["conflicts encountered"],
            "remerge hello --merge-type weave",
            retcode=1,
            working_dir="this",
        )

        self.assertPathExists("this/hello.OTHER")
        self.assertPathDoesNotExist("this/question.OTHER")

        self.run_bzr("file-id hello", working_dir="this")[0]
        self.run_bzr_error(
            ["hello.THIS is not versioned"], "file-id hello.THIS", working_dir="this"
        )

        self.run_bzr_error(
            ["conflicts encountered"],
            "remerge --merge-type weave",
            retcode=1,
            working_dir="this",
        )

        self.assertPathExists("this/hello.OTHER")
        self.assertTrue("this/hello.BASE")
        with open("this/hello") as f:
            conflict_text = f.read()
        self.assertFalse("|||||||" in conflict_text)
        self.assertFalse("hi world" in conflict_text)

        self.run_bzr_error(
            ["Showing base is not supported.*Weave"],
            "remerge . --merge-type weave --show-base",
            working_dir="this",
        )
        self.run_bzr_error(
            ["Can't reprocess and show base"],
            "remerge . --show-base --reprocess",
            working_dir="this",
        )
        self.run_bzr_error(
            ["conflicts encountered"],
            "remerge . --merge-type weave --reprocess",
            retcode=1,
            working_dir="this",
        )
        self.run_bzr_error(
            ["conflicts encountered"],
            "remerge hello --show-base",
            retcode=1,
            working_dir="this",
        )
        self.run_bzr_error(
            ["conflicts encountered"],
            "remerge hello --reprocess",
            retcode=1,
            working_dir="this",
        )

        self.run_bzr("resolve --all", working_dir="this")
        self.run_bzr("commit -m done", working_dir="this")

        self.run_bzr_error(
            [
                "remerge only works after normal merges",
                "Not cherrypicking or multi-merges",
            ],
            "remerge",
            working_dir="this",
        )

    def test_conflicts(self):
        self.create_conflicts()
        self.run_bzr("merge ../other", retcode=1, working_dir="this")
        wt = WorkingTree.open("this")
        self.assertEqual(2, len(wt.conflicts()))
        self.run_bzr("remerge", retcode=1, working_dir="this")
        wt = WorkingTree.open("this")
        self.assertEqual(2, len(wt.conflicts()))
        self.run_bzr("remerge hello", retcode=1, working_dir="this")
        wt = WorkingTree.open("this")
        self.assertEqual(2, len(wt.conflicts()))



# Copyright (C) 2017 The Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test for 'brz mv'."""

from breezy.tests import TestCaseWithTransport


class TestCopy(TestCaseWithTransport):
    def test_cp_unversioned(self):
        self.build_tree(["unversioned.txt"])
        self.run_bzr_error(
            [
                "^brz: ERROR: Could not copy .*unversioned.txt => .*elsewhere."
                " .*unversioned.txt is not versioned\\.$"
            ],
            "cp unversioned.txt elsewhere",
        )

    def test_cp_nonexisting(self):
        self.run_bzr_error(
            [
                "^brz: ERROR: Could not copy .*doesnotexist => .*somewhereelse."
                " .*doesnotexist is not versioned\\.$"
            ],
            "cp doesnotexist somewhereelse",
        )

    def test_cp_unqualified(self):
        self.run_bzr_error(["^brz: ERROR: missing file argument$"], "cp")

    def test_cp_invalid(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["test.txt", "sub1/"])
        tree.add(["test.txt"])

        self.run_bzr_error(
            [
                "^brz: ERROR: Could not copy test.txt => sub1/test.txt: "
                "sub1 is not versioned\\.$"
            ],
            "cp test.txt sub1",
        )

        self.run_bzr_error(
            [
                "^brz: ERROR: Could not copy test.txt => .*hello.txt: "
                "sub1 is not versioned\\.$"
            ],
            "cp test.txt sub1/hello.txt",
        )

    def test_cp_dir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["hello.txt", "sub1/"])
        tree.add(["hello.txt", "sub1"])

        self.run_bzr_error(
            ["^brz: ERROR: Could not copy sub1 => sub2 . sub1 is a directory\\.$"],
            "cp sub1 sub2",
        )

    def test_cp_file_into(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["sub1/", "sub1/hello.txt", "sub2/"])
        tree.add(["sub1", "sub1/hello.txt", "sub2"])

        self.run_bzr("cp sub1/hello.txt sub2")
        self.assertInWorkingTree("sub1")
        self.assertInWorkingTree("sub1/hello.txt")
        self.assertInWorkingTree("sub2")
        self.assertInWorkingTree("sub2/hello.txt")

    def test_cp_file(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["hello.txt"])
        tree.add(["hello.txt"])

        self.run_bzr("cp hello.txt hallo.txt")
        self.assertInWorkingTree("hello.txt")
        self.assertInWorkingTree("hallo.txt")



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for InterBranch.fetch."""

from breezy.errors import FetchLimitUnsupported, NoRoundtrippingSupport
from breezy.revision import NULL_REVISION
from breezy.tests import TestNotApplicable
from breezy.tests.per_interbranch import TestCaseWithInterBranch


class TestInterBranchFetch(TestCaseWithInterBranch):
    def test_fetch_revisions(self):
        """Test fetch-revision operation."""
        wt = self.make_from_branch_and_tree("b1")
        b1 = wt.branch
        self.build_tree_contents([("b1/foo", b"hello")])
        wt.add(["foo"])
        rev1 = wt.commit("lala!", allow_pointless=False)

        b2 = self.make_to_branch("b2")
        try:
            b2.fetch(b1)
        except NoRoundtrippingSupport:
            raise TestNotApplicable(
                "lossless cross-vcs fetch {!r} to {!r} not supported".format(b1, b2)
            )

        # fetch does not update the last revision
        self.assertEqual(NULL_REVISION, b2.last_revision())

        b2.repository.get_revision(rev1)
        tree = b2.repository.revision_tree(rev1)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(tree.get_file_text("foo"), b"hello")

    def test_fetch_revisions_limit(self):
        """Test fetch-revision operation."""
        builder = self.make_branch_builder(
            "b1", format=self.branch_format_from._matchingcontroldir
        )
        builder.start_series()
        rev1 = builder.build_commit()
        rev2 = builder.build_commit()
        rev3 = builder.build_commit()
        builder.finish_series()
        b1 = builder.get_branch()
        b2 = self.make_to_branch("b2")
        try:
            b2.fetch(b1, limit=1)
        except FetchLimitUnsupported:
            raise TestNotApplicable("interbranch does not support fetch limits")
        except NoRoundtrippingSupport:
            raise TestNotApplicable(
                "lossless cross-vcs fetch {!r} to {!r} not supported".format(b1, b2)
            )

        # fetch does not update the last revision
        self.assertEqual(NULL_REVISION, b2.last_revision())

        self.assertEqual({rev1}, b2.repository.has_revisions([rev1, rev2, rev3]))

    def test_fetch_revisions_limit_incremental(self):
        """Test incremental fetch-revision operation with limit."""
        wt = self.make_from_branch_and_tree("b1")
        b1 = wt.branch
        self.build_tree_contents([("b1/foo", b"hello")])
        wt.add(["foo"])
        rev1 = wt.commit("lala!", allow_pointless=False)

        b2 = self.make_to_branch("b2")
        try:
            b2.fetch(b1, limit=1)
        except FetchLimitUnsupported:
            raise TestNotApplicable("interbranch does not support fetch limits")
        except NoRoundtrippingSupport:
            raise TestNotApplicable(
                "lossless cross-vcs fetch {!r} to {!r} not supported".format(b1, b2)
            )

        self.assertEqual(
            {rev1}, b2.repository.has_revisions([rev1, b"revision-2", b"revision-3"])
        )

        rev2 = wt.commit("hmm")
        rev3 = wt.commit("hmmm")

        b2.fetch(b1, limit=1)

        # fetch does not update the last revision
        self.assertEqual(NULL_REVISION, b2.last_revision())

        self.assertEqual({rev1, rev2}, b2.repository.has_revisions([rev1, rev2, rev3]))



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the exporter."""

import gzip
import os
import tempfile

from .... import tests
from ..exporter import _get_output_stream, check_ref_format, sanitize_ref_name_for_git
from . import FastimportFeature


class TestOutputStream(tests.TestCase):
    _test_needs_features = [FastimportFeature]

    def test_get_output_stream_stdout(self):
        # - returns standard out
        self.assertIsNot(None, _get_output_stream("-"))

    def test_get_source_gz(self):
        fd, filename = tempfile.mkstemp(suffix=".gz")
        os.close(fd)
        with _get_output_stream(filename) as stream:
            stream.write(b"bla")
        # files ending in .gz are automatically decompressed.
        with gzip.GzipFile(filename) as f:
            self.assertEqual(b"bla", f.read())

    def test_get_source_file(self):
        # other files are opened as regular files.
        fd, filename = tempfile.mkstemp()
        os.close(fd)
        with _get_output_stream(filename) as stream:
            stream.write(b"foo")
        with open(filename) as f:
            self.assertEqual("foo", f.read())


# from dulwich.tests.test_repository:
class CheckRefFormatTests(tests.TestCase):
    """Tests for the check_ref_format function.

    These are the same tests as in the git test suite.
    """

    def test_valid(self):
        self.assertTrue(check_ref_format(b"heads/foo"))
        self.assertTrue(check_ref_format(b"foo/bar/baz"))
        self.assertTrue(check_ref_format(b"refs///heads/foo"))
        self.assertTrue(check_ref_format(b"foo./bar"))
        self.assertTrue(check_ref_format(b"heads/foo@bar"))
        self.assertTrue(check_ref_format(b"heads/fix.lock.error"))

    def test_invalid(self):
        self.assertFalse(check_ref_format(b"foo"))
        self.assertFalse(check_ref_format(b"foo/.bar"))
        self.assertFalse(check_ref_format(b"heads/foo/"))
        self.assertFalse(check_ref_format(b"heads/foo."))
        self.assertFalse(check_ref_format(b"./foo"))
        self.assertFalse(check_ref_format(b".refs/foo"))
        self.assertFalse(check_ref_format(b"heads/foo..bar"))
        self.assertFalse(check_ref_format(b"heads/foo?bar"))
        self.assertFalse(check_ref_format(b"heads/foo.lock"))
        self.assertFalse(check_ref_format(b"heads/v@{ation"))
        self.assertFalse(check_ref_format(b"heads/foo\\bar"))
        self.assertFalse(check_ref_format(b"heads/foo\bar"))
        self.assertFalse(check_ref_format(b"heads/foo bar"))
        self.assertFalse(check_ref_format(b"heads/foo\020bar"))
        self.assertFalse(check_ref_format(b"heads/foo\177bar"))


class CheckRefnameRewriting(tests.TestCase):
    """Tests for sanitize_ref_name_for_git function."""

    def test_passthrough_valid(self):
        self.assertEqual(sanitize_ref_name_for_git(b"heads/foo"), b"heads/foo")
        self.assertEqual(sanitize_ref_name_for_git(b"foo/bar/baz"), b"foo/bar/baz")
        self.assertEqual(
            sanitize_ref_name_for_git(b"refs///heads/foo"), b"refs///heads/foo"
        )
        self.assertEqual(sanitize_ref_name_for_git(b"foo./bar"), b"foo./bar")
        self.assertEqual(sanitize_ref_name_for_git(b"heads/foo@bar"), b"heads/foo@bar")
        self.assertEqual(
            sanitize_ref_name_for_git(b"heads/fix.lock.error"), b"heads/fix.lock.error"
        )

    def test_rewrite_invalid(self):
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"foo./bar")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/foo/")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/foo.")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"./foo")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b".refs/foo")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/foo..bar")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/foo?bar")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/foo.lock")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/v@{ation")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/foo\bar")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/foo\\bar")))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b"heads/foo bar")))
        self.assertTrue(
            check_ref_format(sanitize_ref_name_for_git(b"heads/foo\020bar"))
        )
        self.assertTrue(
            check_ref_format(sanitize_ref_name_for_git(b"heads/foo\177bar"))
        )



#! /usr/bin/env python3

"""Installation script for brz.
Run it with
 './setup.py install', or
 './setup.py --help' for more options.
"""

import importlib.util
import os
import os.path
import sys

if not importlib.util.find_spec("setuptools"):
    sys.stderr.write("[ERROR] Please install setuptools ({})\n".format(e))
    sys.exit(1)

try:
    from setuptools_rust import Binding, RustExtension, Strip
except ModuleNotFoundError as e:
    sys.stderr.write("[ERROR] Please install setuptools_rust ({})\n".format(e))
    sys.exit(1)


from setuptools import setup

try:
    from packaging.version import Version
except ImportError:
    from distutils.version import LooseVersion as Version

from distutils.command.build_scripts import build_scripts

from setuptools import Command

###############################
# Overridden distutils actions
###############################


class brz_build_scripts(build_scripts):
    """Fixup Rust extension binary files to live under scripts."""

    def run(self):
        build_scripts.run(self)

        self.run_command("build_ext")
        build_ext = self.get_finalized_command("build_ext")

        for ext in self.distribution.rust_extensions:
            if ext.binding == Binding.Exec:
                # GZ 2021-08-19: Not handling multiple binaries yet.
                os.replace(
                    os.path.join(build_ext.build_lib, ext.name),
                    os.path.join(self.build_dir, ext.name),
                )


class build_man(Command):
    """Generate brz.1."""

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from tools import generate_docs

        generate_docs.main(argv=["brz", "man"])


########################
## Setup
########################

from setuptools.command.build import build

build.sub_commands.append(("build_mo", lambda _: True))

command_classes = {
    "build_man": build_man,
}

from distutils.extension import Extension

ext_modules = []
try:
    from Cython.Compiler.Version import version as cython_version
    from Cython.Distutils import build_ext
except ModuleNotFoundError:
    have_cython = False
    # try to build the extension from the prior generated source.
    print("")
    print(
        "The python package 'Cython' is not available. If the .c files are available,"
    )
    print("they will be built, but modifying the .pyx files will not rebuild them.")
    print("")
    from distutils.command.build_ext import build_ext
else:
    minimum_cython_version = "0.29"
    cython_version_info = Version(cython_version)
    if cython_version_info < Version(minimum_cython_version):
        print(
            "Version of Cython is too old. Current is {}, need at least {}.".format(
                cython_version, minimum_cython_version
            )
        )
        print(
            "If the .c files are available, they will be built,"
            " but modifying the .pyx files will not rebuild them."
        )
        have_cython = False
    else:
        have_cython = True


# Override the build_ext if we have Cython available
command_classes["build_ext"] = build_ext
unavailable_files = []


def add_cython_extension(module_name, libraries=None, extra_source=None):
    """Add a cython module to build.

    This will use Cython to auto-generate the .c file if it is available.
    Otherwise it will fall back on the .c file. If the .c file is not
    available, it will warn, and not add anything.

    You can pass any extra options to Extension through kwargs. One example is
    'libraries = []'.

    :param module_name: The python path to the module. This will be used to
        determine the .pyx and .c files to use.
    """
    if extra_source is None:
        extra_source = []
    path = module_name.replace(".", "/")
    cython_name = path + ".pyx"
    c_name = path + ".c"
    define_macros = []
    if sys.platform == "win32":
        # cython uses the macro WIN32 to detect the platform, even though it
        # should be using something like _WIN32 or MS_WINDOWS, oh well, we can
        # give it the right value.
        define_macros.append(("WIN32", None))
    if have_cython:
        source = [cython_name]
    else:
        if not os.path.isfile(c_name):
            unavailable_files.append(c_name)
            return
        else:
            source = [c_name]
    source.extend(extra_source)
    include_dirs = ["breezy"]
    ext_modules.append(
        Extension(
            module_name,
            source,
            define_macros=define_macros,
            libraries=libraries,
            include_dirs=include_dirs,
            optional=os.environ.get("CIBUILDWHEEL", "0") != "1",
        )
    )


add_cython_extension("breezy.bzr._simple_set_pyx")
ext_modules.append(
    Extension("breezy.bzr._static_tuple_c", ["breezy/bzr/_static_tuple_c.c"])
)
add_cython_extension("breezy._annotator_pyx")
add_cython_extension("breezy._chunks_to_lines_pyx")
add_cython_extension(
    "breezy.bzr._groupcompress_pyx", extra_source=["breezy/bzr/diff-delta.c"]
)
add_cython_extension("breezy.bzr._knit_load_data_pyx")
add_cython_extension("breezy._known_graph_pyx")
add_cython_extension("breezy.bzr._rio_pyx")
if sys.platform == "win32":
    add_cython_extension("breezy.bzr._dirstate_helpers_pyx", libraries=["Ws2_32"])
else:
    add_cython_extension("breezy.bzr._dirstate_helpers_pyx")
    add_cython_extension("breezy._readdir_pyx")
add_cython_extension("breezy.bzr._chk_map_pyx")
add_cython_extension("breezy.bzr._btree_serializer_pyx")


if unavailable_files:
    print("C extension(s) not found:")
    print("   {}".format("\n  ".join(unavailable_files)))
    print("The python versions will be used instead.")
    print("")


if "editable_wheel" not in sys.argv:
    command_classes["build_scripts"] = brz_build_scripts


# ad-hoc for easy_install
DATA_FILES = []
if (
    "bdist_egg" not in sys.argv
    and "bdist_wheel" not in sys.argv
    and "editable_wheel" not in sys.argv
):
    # generate and install brz.1 only with plain install, not the
    # easy_install one
    build.sub_commands.append(("build_man", lambda _: True))
    DATA_FILES = [("man/man1", ["brz.1", "breezy/git/git-remote-bzr.1"])]

import site

site.ENABLE_USER_SITE = "--user" in sys.argv

rust_extensions = [
    RustExtension("breezy.bzr._rio_rs", "lib-rio/Cargo.toml", binding=Binding.PyO3),
]
entry_points = {}

if (
    os.environ.get("CIBUILDWHEEL", "0") == "0"
    and "__pypy__" not in sys.builtin_module_names
    and sys.platform != "win32"
):
    rust_extensions.append(RustExtension("brz", binding=Binding.Exec, strip=Strip.All))
else:
    # Fall back to python main on cibuildwheels, since it doesn't provide
    # -lpython3.7 to link binaries against

    # also, disable it for PyPy. See https://foss.heptapod.net/pypy/pypy/-/issues/3286
    entry_points.setdefault("console_scripts", []).append("brz=breezy.__main__:main")

# std setup
setup(
    scripts=[  # TODO(jelmer): Only install the git scripts if
        # Dulwich was found.
        "breezy/git/git-remote-bzr",
        "breezy/git/bzr-receive-pack",
        "breezy/git/bzr-upload-pack",
    ],
    data_files=DATA_FILES,
    cmdclass=command_classes,
    ext_modules=ext_modules,
    entry_points=entry_points,
    rust_extensions=rust_extensions,
)



##############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Bootstrap a buildout-based project.

Simply run this script in a directory containing a buildout.cfg.
The script accepts buildout command-line options, so you can
use the -c option to specify an alternate configuration file.

$Id: bootstrap.py 90478 2008-08-27 22:44:46Z georgyberdyshev $
"""

import os
import shutil
import sys
import tempfile

import urllib2

tmpeggs = tempfile.mkdtemp()

is_jython = sys.platform.startswith("java")

try:
    import pkg_resources
except ModuleNotFoundError:
    ez = {}
    exec(urllib2.urlopen("http://peak.telecommunity.com/dist/ez_setup.py").read(), ez)
    ez["use_setuptools"](to_dir=tmpeggs, download_delay=0)

    import pkg_resources

if sys.platform == "win32":

    def quote(c):
        if " " in c:
            return '"{}"'.format(c)  # work around spawn lamosity on windows
        else:
            return c
else:

    def quote(c):
        return c


cmd = "from setuptools.command.easy_install import main; main()"
ws = pkg_resources.working_set
env = dict(
    os.environ,
    PYTHONPATH=ws.find(pkg_resources.Requirement.parse("setuptools")).location,
)

if is_jython:
    import subprocess

    assert (
        subprocess.Popen(
            [sys.executable]
            + ["-c", quote(cmd), "-mqNxd", quote(tmpeggs), "zc.buildout"],
            env=env,
        ).wait()
        == 0
    )

else:
    assert (
        os.spawnle(
            os.P_WAIT,
            sys.executable,
            quote(sys.executable),
            "-c",
            quote(cmd),
            "-mqNxd",
            quote(tmpeggs),
            "zc.buildout",
            env,
        )
        == 0
    )

ws.add_entry(tmpeggs)
ws.require("zc.buildout")
import zc.buildout.buildout

zc.buildout.buildout.main(sys.argv[1:] + ["bootstrap"])
shutil.rmtree(tmpeggs)



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ...bzr import inventory
from ...bzr.inventory import ROOT_ID, Inventory
from ...bzr.xml_serializer import (
    Element,
    SubElement,
    XMLSerializer,
    escape_invalid_chars,
)
from ...errors import BzrError
from ...revision import Revision


class _Serializer_v4(XMLSerializer):
    """Version 0.0.4 serializer.

    You should use the serializer_v4 singleton.

    v4 serialisation is no longer supported, only deserialisation.
    """

    __slots__: list[str] = []

    def _pack_entry(self, ie):
        """Convert InventoryEntry to XML element."""
        e = Element("entry")
        e.set("name", ie.name)
        e.set("file_id", ie.file_id.decode("ascii"))
        e.set("kind", ie.kind)

        if ie.text_size is not None:
            e.set("text_size", f"{int(ie.text_size)}")

        for f in ["text_id", "text_sha1", "symlink_target"]:
            v = getattr(ie, f)
            if v is not None:
                e.set(f, v)

        # to be conservative, we don't externalize the root pointers
        # for now, leaving them as null in the xml form.  in a future
        # version it will be implied by nested elements.
        if ie.parent_id != ROOT_ID:
            e.set("parent_id", ie.parent_id)

        e.tail = "\n"

        return e

    def _unpack_inventory(
        self, elt, revision_id=None, entry_cache=None, return_from_cache=False
    ):
        """Construct from XML Element.

        :param revision_id: Ignored parameter used by xml5.
        """
        root_id = elt.get("file_id")
        root_id = root_id.encode("ascii") if root_id else ROOT_ID
        inv = Inventory(root_id)
        for e in elt:
            ie = self._unpack_entry(
                e, entry_cache=entry_cache, return_from_cache=return_from_cache
            )
            if ie.parent_id == ROOT_ID:
                ie.parent_id = root_id
            inv.add(ie)
        return inv

    def _unpack_entry(self, elt, entry_cache=None, return_from_cache=False):
        # original format inventories don't have a parent_id for
        # nodes in the root directory, but it's cleaner to use one
        # internally.
        parent_id = elt.get("parent_id")
        parent_id = parent_id.encode("ascii") if parent_id else ROOT_ID

        file_id = elt.get("file_id").encode("ascii")
        kind = elt.get("kind")
        if kind == "directory":
            ie = inventory.InventoryDirectory(file_id, elt.get("name"), parent_id)
        elif kind == "file":
            ie = inventory.InventoryFile(file_id, elt.get("name"), parent_id)
            ie.text_id = elt.get("text_id")
            if ie.text_id is not None:
                ie.text_id = ie.text_id.encode("utf-8")
            ie.text_sha1 = elt.get("text_sha1")
            if ie.text_sha1 is not None:
                ie.text_sha1 = ie.text_sha1.encode("ascii")
            v = elt.get("text_size")
            ie.text_size = v and int(v)
        elif kind == "symlink":
            ie = inventory.InventoryLink(file_id, elt.get("name"), parent_id)
            ie.symlink_target = elt.get("symlink_target")
        else:
            raise BzrError("unknown kind {!r}".format(kind))

        ## mutter("read inventoryentry: %r", elt.attrib)

        return ie

    def _pack_revision(self, rev):
        """Revision object -> xml tree."""
        root = Element(
            "revision",
            committer=rev.committer,
            timestamp="{:.9f}".format(rev.timestamp),
            revision_id=rev.revision_id,
            inventory_id=rev.inventory_id,
            inventory_sha1=rev.inventory_sha1,
        )
        if rev.timezone:
            root.set("timezone", str(rev.timezone))
        root.text = "\n"

        msg = SubElement(root, "message")
        msg.text = escape_invalid_chars(rev.message)[0]
        msg.tail = "\n"

        if rev.parents:
            pelts = SubElement(root, "parents")
            pelts.tail = pelts.text = "\n"
            for i, parent_id in enumerate(rev.parents):
                p = SubElement(pelts, "revision_ref")
                p.tail = "\n"
                p.set("revision_id", parent_id)
                if i < len(rev.parent_sha1s):
                    p.set("revision_sha1", rev.parent_sha1s[i])
        return root

    def write_revision_to_string(self, rev):
        return tostring(self._pack_revision(rev)) + b"\n"

    def _write_element(self, elt, f):
        ElementTree(elt).write(f, "utf-8")
        f.write(b"\n")

    def _unpack_revision(self, elt):
        """XML Element -> Revision object."""
        # <changeset> is deprecated...
        if elt.tag not in ("revision", "changeset"):
            raise BzrError("unexpected tag in revision file: {!r}".format(elt))

        rev = Revision(
            committer=elt.get("committer"),
            timestamp=float(elt.get("timestamp")),
            revision_id=elt.get("revision_id"),
            inventory_id=elt.get("inventory_id"),
            inventory_sha1=elt.get("inventory_sha1"),
        )

        precursor = elt.get("precursor")
        precursor_sha1 = elt.get("precursor_sha1")

        pelts = elt.find("parents")

        if pelts is not None:
            for p in pelts:
                rev.parent_ids.append(p.get("revision_id"))
                rev.parent_sha1s.append(p.get("revision_sha1"))
            if precursor:
                # must be consistent
                rev.parent_ids[0]
        elif precursor:
            # revisions written prior to 0.0.5 have a single precursor
            # give as an attribute
            rev.parent_ids.append(precursor)
            rev.parent_sha1s.append(precursor_sha1)

        v = elt.get("timezone")
        rev.timezone = v and int(v)

        rev.message = elt.findtext("message")  # text of <message>
        return rev


"""singleton instance"""
serializer_v4 = _Serializer_v4()
